//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "RegExtension.h"
#pragma package(smart_init)

#define WM_SHELLNOTIFY  WM_USER
//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//

static inline void ValidCtrCheck(TRegExtension *)
{
    new TRegExtension(NULL);
}
//---------------------------------------------------------------------------
__fastcall TRegExtension::TRegExtension(TComponent* Owner)
    : TComponent(Owner), hNotifyHandle(0), isNT(Win32Platform == VER_PLATFORM_WIN32_NT)

{
    Reg = new TRegistry;
    Reg->RootKey = HKEY_CLASSES_ROOT;

    FRegExts = new TStringList;
    FRegExts->Sorted = true;
    FRegExts->Clear();
    FUpdateExts = true;

    FExtensions = new TList;
    FExtensions->Clear();
}
//---------------------------------------------------------------------------
__fastcall TRegExtension::~TRegExtension()
{
    Clear();
    delete FExtensions;
    delete FRegExts;
    delete Reg;

    StopWatching();
}
//---------------------------------------------------------------------------
int __fastcall TRegExtension::Add(TExtInfo *E)
{
    if(E)
        return FExtensions->Add(E);
    else
        return -1;
}
//---------------------------------------------------------------------------
int __fastcall TRegExtension::Add(AnsiString Ext, AnsiString ExecPath)
{
    return Add(Ext, ExecPath, (Ext[1]=='.') ? (Ext.SubString(2, Ext.Length()) + " File") : (Ext + " File"), 0);
}
//---------------------------------------------------------------------------
int __fastcall TRegExtension::Add(AnsiString Ext, AnsiString ExecPath, AnsiString Desc)
{
    return Add(Ext, ExecPath, Desc, 0);
}
//---------------------------------------------------------------------------
int __fastcall TRegExtension::Add(AnsiString Ext, AnsiString ExecPath, AnsiString Desc, DWORD Flags)
{
    return Add(new TExtInfo(Ext, ExecPath, Desc, Flags));
}
//---------------------------------------------------------------------------
void __fastcall TRegExtension::AssocChanged(void)
{
    FUpdateExts = true;	
    if(FOnAssocChange)
        FOnAssocChange(this);
}
//---------------------------------------------------------------------------
void __fastcall TRegExtension::Clear(void)
{
    for(int x = 0; x < FExtensions->Count; x++)
        delete (TExtInfo*)FExtensions->Items[x];
    FExtensions->Clear();
}
//---------------------------------------------------------------------------
bool __fastcall TRegExtension::CreateExtension(TExtInfo *E)
{
    if((!E->Extension.IsEmpty()) && (!E->ExecutePath.IsEmpty()))
    {
        AnsiString ext, name, basekeyname, keyname, basecommandskey;

        ext = E->Extension.SubString(2, E->Extension.Length());
        name = (ext + "file");

        if(WriteValue(("\\" + E->Extension), DEFAULT, name))
        {
            basekeyname = ("\\" + name);
            if(WriteValue(basekeyname, DEFAULT, E->FileDesc))
            {
                DWORD flags = E->EditFlags;
                if(WriteBinary(basekeyname, "EditFlags", &flags, sizeof(DWORD)))
                {
                    keyname = (basekeyname + "\\DefaultIcon");
                    if(WriteValue(keyname, DEFAULT, E->Icon))
                    {
                        if((E->Commands->Count == 0) && E->Commands->AutoCreate)
                            E->Commands->Add(EXT_OPEN, "&Open", "\"%1\"");

                        basecommandskey = (basekeyname + "\\shell");

                        for(int x = 0; x < E->Commands->Count; x++)
                        {
                            TExtCommand *C = E->Commands->Command[x];
                            if(C != NULL)
                            {
                                if(WriteValue(basecommandskey, DEFAULT, E->Commands->DefaultAction))
                                {
                                    keyname = (basecommandskey + "\\" + C->Action);
                                    if(WriteValue(keyname, DEFAULT, C->ActionDesc))
                                    {
                                        keyname = (keyname + "\\command");
                                        WriteValue(keyname, DEFAULT, "\"" + E->ExecutePath + "\" " + C->Command);
                                    }
                                }
                            }
                        }

                        return true;
                    }
                }
            }
        }
    }

    return false;
}
//---------------------------------------------------------------------------
bool __fastcall TRegExtension::DeleteKey(AnsiString key)
{
    if(Reg->KeyExists(key))
    {
        bool dodelete = true;
        if(Win32Platform == VER_PLATFORM_WIN32_NT)
        {
            TStringList *subkeys = new TStringList;
            GetSubkeys(key, subkeys);
            int count = subkeys->Count;
            for(int x = 0; x < count && dodelete; x++)
                dodelete = DeleteKey(key + "\\" + subkeys->Strings[x]);
            delete subkeys;
        }
        if(!dodelete)
            return false;
        return Reg->DeleteKey(key);
    }
    return false;
}
//---------------------------------------------------------------------------
TExtInfo* __fastcall TRegExtension::GetExt(int index)
{
    if((index >= 0) && (index < FExtensions->Count))
        return (TExtInfo*)FExtensions->Items[index];
    else
        return NULL;
}
//---------------------------------------------------------------------------
TExtInfo* __fastcall TRegExtension::GetExtInfo(AnsiString Ext)
{
    TExtInfo *info = NULL;

    if(Ext[1] != '.')
        Ext = ("." + Ext);

    if(IsRegistered(Ext))
    {
        AnsiString name = ReadValue(("\\" + Ext), DEFAULT);

        if(!name.IsEmpty())
        {
            DWORD editflags = 0;
            ReadBinary("\\" + name, "EditFlags", &editflags, sizeof(DWORD));

            AnsiString desc = ReadValue(("\\" + name), DEFAULT);
            AnsiString execpath = ReadValue(("\\" + name + "\\shell\\open\\command"), DEFAULT);

            info = new TExtInfo(Ext, execpath, desc, editflags);
            info->Icon = ReadValue(("\\" + name + "\\DefaultIcon"), DEFAULT);

            info->Commands->DefaultAction = ReadValue(("\\" + name + "\\shell"), DEFAULT);

            TStringList *subkeys = new TStringList;
            GetSubkeys("\\" + name + "\\shell", subkeys);
            for(int x = 0; x < subkeys->Count; x++)
            {
                AnsiString action, actiondesc, command, key;

                action = subkeys->Strings[x];
                key = ("\\" + name + "\\shell\\" + action);

                actiondesc = ReadValue(key, DEFAULT);
                command = ReadValue((key + "\\command"), DEFAULT);

                info->Commands->Add(action, actiondesc, command);
            }
            delete subkeys;
        }
    }

    return info;
}
//---------------------------------------------------------------------------
TStringList* __fastcall TRegExtension::GetRegExts(void)
{
    UpdateRegExts(FUpdateExts);
    return FRegExts;
}
//---------------------------------------------------------------------------
void __fastcall TRegExtension::GetSubkeys(AnsiString key, TStringList *SubKeys)
{
    SubKeys->Clear();

    if(!key.IsEmpty())
    {
        if(Reg->OpenKey(key, false))
        {
            Reg->GetKeyNames(SubKeys);
            Reg->CloseKey();
        }
    }
}
//---------------------------------------------------------------------------
void __fastcall TRegExtension::HandleMessage(TMessage &Message)
{
    DWORD EventId = 0;

    // Handle only the WM_SHELLNOTIFY message
    if(Message.Msg == WM_SHELLNOTIFY)
    {
        // If this is NT, use the memory map to access the PIDL data
        if(isNT)
        {
            TTwoPIDLArray PIDLs;
            HANDLE Lock = SHChangeNotification_Lock((HANDLE)Message.WParam, (DWORD)Message.LParam, &PIDLs, EventId);
            if(Lock != NULL)
            {
                if(EventId & SHCNE_ASSOCCHANGED)
                    AssocChanged();
                SHChangeNotification_Unlock(Lock);
            }
        }
        // If this is not NT, access the PIDL data directly. }
        else
        {
            EventId = (DWORD)Message.LParam;
            if(EventId & SHCNE_ASSOCCHANGED)
                AssocChanged();
        }
    }
    // Call the default windows proc for any other message. }
    else
        Message.Result = DefWindowProc(hWnd, Message.Msg, Message.WParam, Message.LParam);
}
//---------------------------------------------------------------------------
bool __fastcall TRegExtension::IsRegistered(AnsiString Ext)
{
    return Reg->KeyExists(("\\" + (Ext[1]=='.')?Ext:("." + Ext)));
}
//---------------------------------------------------------------------------
bool __fastcall TRegExtension::ReadBinary(AnsiString key, AnsiString name, void *data, int size)
{
    bool retVal = false;

    if(Reg->OpenKey(key, false))
    {
        try
        {
            retVal = Reg->ReadBinaryData(name, data, size);
        }
        catch(...)
        {
            retVal = false;
        }
        Reg->CloseKey();
    }

    return retVal;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TRegExtension::ReadValue(AnsiString key, AnsiString name)
{
    AnsiString retVal;

    if(Reg->OpenKey(key, false))
    {
        retVal = Reg->ReadString(name);
        Reg->CloseKey();
    }

    return retVal;
}
//---------------------------------------------------------------------------
bool __fastcall TRegExtension::Register(void)
{
    bool retval = false;
    for(int x = 0; x < FExtensions->Count; x++)
    {
        retval = CreateExtension((TExtInfo*)FExtensions->Items[x]);
        if(retval == false)
            break;
    }

    SHChangeNotify(SHCNE_ASSOCCHANGED, SHCNF_IDLIST, 0, 0);
    FUpdateExts = true;
    return retval;
}
//---------------------------------------------------------------------------
void __fastcall TRegExtension::SetMonitor(bool value)
{
    if(FMonitor != value)
    {
        FMonitor = value;
        if(FMonitor)
            StartWatching();
        else
            StopWatching();
    }
}
//---------------------------------------------------------------------------
void __fastcall TRegExtension::StartWatching(void)
{
    StopWatching();

    hDll = LoadLibrary("SHELL32.dll");
    if(hDll != NULL)
    {
        hWnd = AllocateHWnd(HandleMessage);

        SHChangeNotifyRegister = (fpSHChangeNotifyRegister)GetProcAddress(hDll, MAKEINTRESOURCE(2));
        SHChangeNotifyDeregister = (fpSHChangeNotifyDeregister)GetProcAddress(hDll, MAKEINTRESOURCE(4));

        if(isNT)
        {
            SHChangeNotification_Lock = (fpSHChangeNotification_Lock)GetProcAddress(hDll, MAKEINTRESOURCE(644));
            SHChangeNotification_Unlock = (fpSHChangeNotification_Unlock)GetProcAddress(hDll, MAKEINTRESOURCE(645));
        }

        if(SHChangeNotifyRegister && SHChangeNotifyDeregister)
        {
            if(isNT && (!SHChangeNotification_Lock || !SHChangeNotification_Unlock))
                return;

            // Initialize Notification Path data
            TNotifyRegister NotifyPathData = {NULL, FALSE};

            // Register for notification and store the handle
            hNotifyHandle = SHChangeNotifyRegister(hWnd, 0x0008/*SHCNF_NO_PROXY*/, SHCNE_ASSOCCHANGED, WM_SHELLNOTIFY, 1, &NotifyPathData);
        }
    }
}
//---------------------------------------------------------------------------
void __fastcall TRegExtension::StopWatching(void)
{
    // Deregister the notification handle
    if(hNotifyHandle && SHChangeNotifyDeregister)
    {
        SHChangeNotifyDeregister(hNotifyHandle);
        DeallocateHWnd(hWnd);
        FreeLibrary(hDll);

        hNotifyHandle = NULL;
        hWnd = NULL;
        hDll = NULL;
    }
}
//---------------------------------------------------------------------------
bool __fastcall TRegExtension::Unregister(System::AnsiString Ext)
{
    bool retval = false;

    if(Ext[1] != '.')
        Ext = ("." + Ext);

    if(IsRegistered(Ext))
    {
        AnsiString key = ReadValue("\\" + Ext, DEFAULT);
        retval = DeleteKey(Ext);
        if(key != "") retval = DeleteKey("\\" + key);
    }

    SHChangeNotify(SHCNE_ASSOCCHANGED, SHCNF_IDLIST, 0, 0);
    FUpdateExts = true;
    return retval;
}
//---------------------------------------------------------------------------
void __fastcall TRegExtension::UpdateRegExts(bool DoUpdate)
{
    if(DoUpdate)
    {
        FRegExts->Clear();
        TStringList *temp = new TStringList;

        if(Reg->OpenKey("\\", false))
        {
            Reg->GetKeyNames(temp);
            Reg->CloseKey();
        }

        for(int x = 0; x < temp->Count; x++)
        {
            AnsiString ext = temp->Strings[x];
            if(ext[1] == '.' && ext.Length() > 1)
                FRegExts->Add(temp->Strings[x]);
        }
        delete temp;

        FUpdateExts = false;
    }
}
//---------------------------------------------------------------------------
bool __fastcall TRegExtension::WriteBinary(AnsiString key, AnsiString name, void *data, int size)
{
    bool retVal = false;

    if(Reg->OpenKey(key, true))
    {
        try
        {
            Reg->WriteBinaryData(name, data, size);
            retVal = true;
        }
        catch (...)
        {
            retVal = false;
        }

        Reg->CloseKey();
    }

    return retVal;
}
//---------------------------------------------------------------------------
bool __fastcall TRegExtension::WriteValue(AnsiString key, AnsiString name, AnsiString value)
{
    bool retVal = false;

    if(Reg->OpenKey(key, true))
    {
        try
        {
            Reg->WriteString(name, value);
            retVal = true;
        }
        catch (...)
        {
            retVal = false;
        }

        Reg->CloseKey();
    }

    return retVal;
}
//---------------------------------------------------------------------------
__fastcall TExtInfo::TExtInfo(AnsiString Ext, AnsiString ExecPath, AnsiString Desc, DWORD Flags)
    : FDesc(Desc), FEditFlags(Flags)
{
    FCommands = new TExtCommands;
    SetExt(Ext);
    SetPath(ExecPath);
}
//---------------------------------------------------------------------------
__fastcall TExtInfo::~TExtInfo(void)
{
    delete FCommands;
}
//---------------------------------------------------------------------------
void __fastcall TExtInfo::SetExt(AnsiString NewExt)
{
    FExt = NewExt;
    if(FDesc.IsEmpty())
        FDesc = (FExt[1]=='.') ? (FExt.SubString(2, FExt.Length()) + " File") : (NewExt + " File");
}
//---------------------------------------------------------------------------
void __fastcall TExtInfo::SetPath(AnsiString NewPath)
{
    FPath = NewPath;
    if(FIcon.IsEmpty() && !FPath.IsEmpty())
        FIcon = (FPath + ",0");
}
//---------------------------------------------------------------------------
__fastcall TExtCommand::TExtCommand(AnsiString Action, AnsiString ActionDesc, AnsiString Command)
{
    FAction = Action;
    FActionDesc = ActionDesc;
    FCommand = (Command=="%1") ? AnsiString("\"%1\"") : Command;
    if(FCommand.AnsiPos("%1")==0)
        FCommand += " \"%1\"";
}
//---------------------------------------------------------------------------
__fastcall TExtCommands::TExtCommands(void)
{
    FCreate = true;
    FActions = new TList;
    FActions->Clear();
}
//---------------------------------------------------------------------------
__fastcall TExtCommands::~TExtCommands(void)
{
    Clear();
    delete FActions;
}
//---------------------------------------------------------------------------
int __fastcall TExtCommands::Add(TExtCommand *C)
{
    if(C)
        return FActions->Add(C);
    else
        return -1;
}
//---------------------------------------------------------------------------
int __fastcall TExtCommands::Add(AnsiString Action, AnsiString ActionDesc)
{
    return Add(Action, ActionDesc, "%1");
}
//---------------------------------------------------------------------------
int __fastcall TExtCommands::Add(AnsiString Action, AnsiString ActionDesc, AnsiString Command)
{
    if(FDefAction.IsEmpty())
        FDefAction = Action;

    return Add(new TExtCommand(Action, ActionDesc, Command));
}
//---------------------------------------------------------------------------
void __fastcall TExtCommands::Clear(void)
{
    for(int x = 0; x < FActions->Count; x++)
        delete (TExtCommand*)FActions->Items[x];
    FActions->Clear();
}
//---------------------------------------------------------------------------
TExtCommand* __fastcall TExtCommands::GetCommand(int index)
{
    if((index >= 0) && (index < FActions->Count))
        return (TExtCommand*)FActions->Items[index];
    else
        return NULL;
}
//---------------------------------------------------------------------------
int __fastcall TExtCommands::GetCount(void)
{
    return FActions->Count;
}
//---------------------------------------------------------------------------

namespace Regextension
{
    void __fastcall PACKAGE Register()
    {
        TComponentClass classes[1] = {__classid(TRegExtension)};
        RegisterComponents("Lebeau", classes, 0);
    }
}
//---------------------------------------------------------------------------
